require "Warshall.rb"
require "Grammatiken.rb"

# Author::    Jens Bachmann  (mailto:jens.bachmann@unibw.de)
# Author::		Daniel Borkowitz (mailto:daniel.borkowitz@unibw.de)
# Author::		Stefan Mayer (mailto:stefan.mayer@unibw.de)
# Author::		Roman Krhling (mailto:roman.kraehling@unibw.de)
# Copyright:: Copyright (c) 2006 The Heimschlfers, UniBw Mnchen

# Diese Klasse stellt ein Item eines Item-Automaten dar, z.B. [S->X.Ya]. (S.154), bzw fr LR(1) usw. noch zustzlich die follow-Menge
class Item
	
	def initialize(string,follow)
	@item=string
	@followmenge= follow
	end
	
	def getItem
	 return @item
	end
	
	def +(it)
	  Item.new(@item + it.getItem, @followmenge)
	
	end
	
	def each &b
		@item.each &b
	end
	
  def copy
	 return Item.new(@items.copy, @followmenge.copy)
	end

	
	# setzt den Punkt um ein Zeichen nach rechts weiter
	def lesen
		ind = getItem.index('.')
		ch = @item[ind+1..ind+1]
		item= Item.new("",@followmenge)
		item+= Item.new(@item[0..ind-1]+ch+"."+@item[ind+2..@item.length],@followmenge)
	end 
 
	def <=>(anOther)
		@item<=>anOther.getItem
	end
	
	def ==(anOther)
		
		return @item == anOther.getItem if anOther!=nil
		return false
		
	end
	
	def sort!()
	@item.sort!
	end
	
	def to_s
		return "[" + @item + "]" if @followmenge==nil
		return "[" + @item + ";" + @followmenge.to_s + "]"
	end
	
	# liefert das nchste Zeichen nach dem Punkt
	def naechstesZeichen
		ind = getItem.index('.')
		return @item[ind+1..ind+1]  
		
	end
	
	# prft ob das Folgezeichen des Punktes ein NonTerminal ist
	def isEpsUebergang(gram)
		return gram.getNonterminals.include?(naechstesZeichen)
	end
	
	# prft ob der Zustand der Ausgangszustand zu eine NonTerminal a ist
	def ist_start_von(a)
		ind = getItem.index('.')		
		return ((@item[ind-2,1]) == ">" and (@item[0,1] == a))
	end
	
end

# Diese Klasse stellt das Tripel einer LR-Konfiguration dar (S.155)
class Konfiguration
	def initialize(zustand, symbol, folge)
		@zustand = zustand
		@symbol = symbol
		@folge = folge
	end
	
	# getter fr die Komponenten des Tripels(0-2)
	def [](i)
	 case i
	    when 0 then @zustand
			when 1 then @symbol
			when 2 then @folge
			else nil
		end
	end
	
	def ==(anOther)
	  if anOther!=nil
			return  (@zustand==anOther[0] and @symbol==anOther[1] and @folge==anOther[2]) 
		else return false
		end
	end
	
	def <=>(anOther)
		return @zustand<=>anOther[0]
	end
	
	def to_s
		return ("[" + @zustand.to_s + " --" + @symbol.to_s + "--> " + @folge.to_s + "]" + "\n")
	end
	
	# liefert den Folgezustand eines Zustands nach "lesen" eines Symbols
	def folgeZustand?(zustand,symbol)
	 if ((zustand==@zustand) and (@symbol == symbol))
		tmp= @folge
	 return tmp
	 else return nil
	 end
	end	
end

# Beschreibt einen Zustand des Item-Automaten, also eine Menge von Items
class Zustand

	def initialize (item)
	@items = Menge.new([])
	@items.add(item) if item != nil
	end
	
	def copy
			temp=Zustand.new(@items[0])   
			@items.each{|i| temp.add_i(i)}
			return temp
	end
	
	def add_i(item)
		@items.add(item) if item!=nil
		return self
	end
	
	def add_Z(zustand)
		@items+=zustand.items if zustand !=nil
		return self
	end
	
	def to_s
		@items.to_s
	end
	
	def items
		@items
	end
	
	def getNEA_item
		@items[0]
	end

	

	
	def ==(another)
		if another!=nil then return (@items == another.items) else return false end
	end
	
	def <=>(anOther)
		@items <=> anOther.items
	end
	
	def sort!
		@items.sort!
	end
	
	def folgeZustand(symbol, uebergang)
		temp=Zustand.new(nil)
	    each{|i| 
			   uebergang.each{ |u| 
				 folge = u.folgeZustand?(Zustand.new(i), symbol)
				 temp.add_Z(folge) if folge!=nil
				 }
				 }
			temp.each{|t|
				 uebergang.each{ |u| 
				 folge = u.folgeZustand?(Zustand.new(t), Symb.new("eps"))
				 temp.add_Z(folge) if folge!=nil
				 }
			}
		return temp	
	end
	
	def each &b
		@items.each &b
	end
end

# Diese Klasse stellt einen Nichtdeterministischen Item-Automaten zu einer Grammatik dar.
class NEA
	
	# der NEA besteht aus einer Menge von Zustnden, einem Startzustand und einer Menge von 
	# Konfigurationen als bergangsfunktion.
	def initialize(grammatik)
			@zustaende = Menge.new([])
			@alphabet= (grammatik.getNonterminals.update(grammatik.getTerminals))
			@startZustand = Zustand.new(Item.new("@ -> .S",nil)) 
			berechneZustaende(grammatik.addRule("@ -> S \n\n"))
			@uebergaenge = uebergangsfunktion(grammatik.addRule("@ -> S \n\n"))
			
 end

	
	# berechnet die Zustnde des Automaten als Menge
	def berechneZustaende(grammatik)
		grammatik.getRules.each_value do |x| 
				links = x.getSymbol
				tmpMenge = Menge.new(x.getAlts)
				tmpMenge.each do |i| 
					for b in 0..i.getSymbols.length do
						str=Item.new(links.to_s+" -> ",nil)
						for c in 0..i.getSymbols.length-1 do
						  
							str+=Item.new(".",nil) if ((c==b))
							str+=Item.new(i.getSymbols[c].to_s,nil)
						end
						str+=Item.new(".",nil) if (b==i.getSymbols.length)
						temp = Zustand.new(str)
						@zustaende.add(temp)
				end
			end
		end
	end
	
	# berechnet die bergnge des Automaten als Menge von LR-Konfigurationen
	def uebergangsfunktion(grammatik)
		erg = Menge.new([])
		@zustaende.each do |z| 
				ch = z.getNEA_item.naechstesZeichen
				 if (z.getNEA_item.isEpsUebergang(grammatik)) then
						@zustaende.each do |folg| 
						erg.add(Konfiguration.new(z,Symb.new("eps"),folg)) if  folg.getNEA_item.ist_start_von(ch.to_s) and folg!=z
						end
					end
					erg.add(Konfiguration.new(z,Symb.new(ch),Zustand.new(z.getNEA_item.lesen))) if ch!=""
			   end
  		return erg
	end
	
	def getUebergaenge
		@uebergaenge
	end
	
	def getZustaende
	 @zustaende
	end
	
	def getStartZustand
		@startZustand
	end
	
	
	def setUebergaenge(ueb)
		@uebergaenge=ueb
	end
	
	def setZustaende(zust)
	 @zustaende=zust
	end
	
	def setStartZustand(st)
		@startZustand=st
	end
	
# Liefert eine wahre Kopie des NEA
	def copy
		temp= NEA.new(Grammar.new(""))
		temp.setUebergaenge(@uebergaenge.copy)
		temp.setZustaende( @zustaende.copy)
		temp.setStartZustand(@startZustand.copy)
		return temp
 end
 # Speichert das Objekt NEA auf der Harddisk mit Dateiname als Parameter mittels Marshall	
 def speichern(dateiname)
		File.open(dateiname, "w") do |f|
			Marshal.dump(self,f)
		end
	end
	
	# Ldt das Objekt NEA von der Harddisk mit Dateiname als Parameter mittels Marshall	
	def laden(dateiname)
		File.open(dateiname, "r") do |f|
			Marshal.load(f)
		end
	end
	
	# Formatierte Ausgabe des Automaten als String
	def to_s
		str= "Startzustand: " +@startZustand.to_s + "\n\n"
		str+= "Zustaende:\n "
				@zustaende.each{|i| str+=("\t\t"+i.to_s+"\n")}
		str+= "\n\n Uebergaenge:\n "
			  @uebergaenge.each{|i| str+= ("\t\t"+i.to_s+"\n")}
		str+="\n\n"
		return str
	end
		

	# each-Iterator fr @zustaende
	def zustaende_each &b
	 @zustaende.each &b
	end
	
	
	
	# each-Iterator fr @uebergaenge
	def uebergaenge_each &b
	 @uebergaenge.each &b
	end
	
end

class LR_0<NEA

	def initialize(grammatik)
		super(grammatik)
		myhill
		uebergaenge
		
	end

	# Myhillsche Teilmengenkonstruktion des zugehrigen DEA S.157
	def myhill
		start = @startZustand.copy
		zustaende = @zustaende.copy
		uebergaenge = @uebergaenge.copy
		@zustaende = Menge.new([])
		@startZustand = Zustand.new(nil)
		@startZustand.add_Z(start)
		@startZustand.add_Z(start.folgeZustand(Symb.new("eps"),uebergaenge))
		@zustaende.add(@startZustand)
  #  puts @startZustand.folgeZustand(Symb.new("b"),uebergaenge).folgeZustand(Symb.new("S"),uebergaenge).folgeZustand(Symb.new("c"),uebergaenge).folgeZustand(Symb.new("S"),uebergaenge).folgeZustand(Symb.new("d"),uebergaenge)
		tmpZustand =Menge.new([])

	  
		while tmpZustand!=@zustaende do
			   tmpZustand=@zustaende.copy
				 tmpZustand.each{|i| 
						@alphabet.each_value{ |z| 
									folge = i.folgeZustand(z,uebergaenge)
									@zustaende.add(folge) if folge !=Zustand.new(nil)
										}
					}
		end
		
	end

	def uebergaenge()
			tempueb= @uebergaenge.copy
			@uebergaenge=Menge.new([])
			@zustaende.each{ |z|
				@alphabet.each_value{ |a|
					folge=z.folgeZustand(a,tempueb)
					@uebergaenge.add(Konfiguration.new(z, a, folge))if folge !=Zustand.new(nil) if a!=Symb.new("eps")
				 }
			}
	end
end
