# Dieses Paket stellt eine Implemtierung eines Syntaxbaumes zur Verfgung.
# Es besteht aus den Klassen Syntaxbaum, der den Start eines Baumes markiert, Knoten, welche einen Knoten
# darstellt, Terminal, welches ein terminaler Knoten ist, und Nonterminal, welcher ein Nonterminaler Knoten ist.
#
# Author:: David Hackbarth

# Die Klasse Knoten stellt einen einzelnen Knoten eines Baumes dar. Jeder Knoten hat einen Namen und einen
# Vater, sowie das Attribute ererbte und synthetisierte, die jeweils als Hash vorliegen.
class Knoten
  
  def initialize(name,vater)
    @name = name
    @vater = vater
    @ererbte = Hash.new()
    @synthetisierte = Hash.new()
  end

# Gibt den Namen des Knoten als String zurck.  
  def to_s(n)
    ausgabe = @name
    return ausgabe
  end
  
# Gibt den Namen des Knoten zurck.
  def getName()
    return @name
  end
    
# Setzt den Namen des Knoten.
  def setName(name)
    @name = name
  end
  
# Gibt den Vater des Knoten zurck.
  def getVater()
    return @vater
  end
  
# Setzt den Vater des Knoten.
  def setVater(vater)
    @vater = vater
  end
  
# Gibt die ererbten Attribute als Hash zurck.
  def getErerbte()
    return @ererbte
  end
  
# Setzt das Hash der ererbten Attribute.
  def setErerbte(ererbte)
    @ererbte = ererbte
  end
  
# Gibt die synthetisierten Attribute als Hash zurck.
  def getSynthetisierte()
    return @synthetisierte
  end

# Setzt das Hash der synthetisierten Attribute.
  def setSynthetisierte(synthetisierte)
    @synthetisierte = synthetisierte
  end
  
# Addiert ein ererbtes Attribut, in der Form <attributname> => <attributwert>, zum Hash.
  def addErerbtes(attributname,attributwert)
    @ererbte[attributname] = attributwert
  end

# Addiert ein synthetisiertes Attribut, in der Form <attributname> => <attributwert>, zum Hash.
  def addSynthetisiertes(attributname,attributwert)
    @synthetisierte[attributname] = attributwert
  end

end

######

# Die Klasse Terminal stammt von der Klasse Knoten ab und stellt ein terminales Ende dar. Es hat
# zustzlich das Attribut Quellstring, um den Bezeichner darstellen zu knnen
class Terminal < Knoten

  def initialize(name,vater,quellstring)
    super(name,vater)
    @quellstring = quellstring
  end

# Gibt den Quellstring des terminalen Knotens zurck.
  def to_s(n)
    ausgabe = @quellstring + "\n"
    return ausgabe
  end
  
# Gibt den Quellstring des Terminals zurck.
  def getQuellstring()
    return @quellstring
  end
  
# Setzt den Quellstring des Terminals.
  def setQuellstring(quellstring)
    @quellstring = quellstring
  end
  
end

######

# Die Klasse Nonterminal stammt von der Klasse Knoten ab und stellt einen nicht terminalen Knoten des Baumes
# dar. Es hat zustzlich das Attribut Shne, um die Nachfolger indentifizieren zu knnen, und das Attribut
# Alternativnr, um zu zeigen um welche Alternative der Regel es sich handelt.
# Shne ist ein Array und besteht aus Nonterminalen oder Terminalen.
class Nonterminal < Knoten

  def initialize(name,vater,alternativnr)
    super(name,vater)
    @alternativnr = alternativnr
    @soehne = Array.new()
  end

# Gibt den gesamten Ast ab diesem Nonterminal als String zurck.
  def to_s(n)
    ausgabe = ""
    @soehne.each do |sohn|
      if sohn != nil
        ausgabe += " "*(n+1) + sohn.to_s(n+1)
      end
    end
    ausgabe  = super + "\n" + ausgabe
    return ausgabe
  end

# Fgt dem Nonterminal ein Nonterminal, eine Stufe tiefer, hinzu.
  def addNonterminal(name,alternativnr)  
    @soehne.push(Nonterminal.new(name,self,alternativnr))
  end
  
# Fgt dem Nonterminal ein Terminal, eine Stufe tiefer, hinzu.
  def addTerminal(name,quellstring)
    @soehne.push(Terminal.new(name,self,quellstring))
  end

# Druckt den gesamten Baum ab diesem Knoten auf der Konsole aus. Die Stufen sind immer ein weiteres
# Leerzeichen eingerckt, um so tiefer die Stufe ist.  
  def drucken()
    print(to_s(0))
  end
  
# Gibt die Alternativnr. des Nonterminals zurck.
  def getAlternativnr()
    return @alternativnr
  end

# Setzt die Alternativnr. des Nonterminals.
  def setAlternativnr(alternativnr)
    @alternativnr = alternativnr
  end
  
# Gibt das Array der Shne des Nonterminals zurck.
  def getSoehne()
    return @soehne
  end
  
# Setzt das Array der Shne des Nonterminals.
  def setSoehne(soehne)
    @soehne = soehne
  end
  
end

######

# Die Klasse Syntaxbaum stammt von der Klasse Nonterminale ab und symbolisiert den ersten Knoten 
# eines Syntaxbaumes.
class Syntaxbaum < Nonterminal

  def initialize(name,alternativnr)
    super(name,nil,alternativnr)
  end
  
end

######

# Diese Methode gibt einen Beispielbaum auf der Konsole aus.
# Ausgabe sollte wie nachfolgend aussehen:
#1
# Kapitel 1.1
# 1.2
#  Kapitel 1.2.1
# 1.3
#  Kapitel 1.3.1
#  1.3.2
#   Kapitel 1.3.2.1
# Kapitel 1.4
# 1.5
#  Kapitel 1.5.1

def bsp()
	a = Syntaxbaum.new("1",0)
	a.addTerminal("1.1","Kapitel 1.1")
	a.addNonterminal("1.2",1)
	a.addNonterminal("1.3",2)
	a.addTerminal("1.4","Kapitel 1.4")
	a.addNonterminal("1.5",0)
	a.getSoehne()[1].addTerminal("1.2.1","Kapitel 1.2.1")
	a.getSoehne()[2].addTerminal("1.3.1","Kapitel 1.3.1")
	a.getSoehne()[2].addNonterminal("1.3.2",1)
	a.getSoehne()[2].getSoehne()[1].addTerminal("1.3.2.1","Kapitel 1.3.2.1")
	a.getSoehne()[4].addTerminal("1.5.1","Kapitel 1.5.1")
	a.drucken
end

# bsp()