# Grammatiken
# ===========
#
# (Vorlufige Version)
# Dieses Programm liest einen Regelteil einer SIC-Grammatik ein
# zerlegt ihn in seine Bestandteile und stellt einige Methoden 
# zu Verfgung um mit den so entstandenen Teilen zu arbeiten.

# Diese Klasse reprsentiert die Symbole der Grammatik ohne zu 
# unterscheiden, ob diese Terminale oder Nicht-Terminale sind.
# Symbole kennen nur ihren Namen und knnen diesen zurckgeben.
class Symb

	# Einfacher Konstruktor
	def initialize(name)
		@name = name
	end
	
	# Einfacher Getter fr @name-Attribut
	def getName
		return @name
	end
end

# Diese Klasse reprsentiert die einzelnen Alternative auf der
# rechten Seite einer Regel. Sie besteht aus einem Array von 
# Symbolen.
class Alternative

	# Der Konstruktor zerlegt einen bergebenen String einer Rechten
	# Seite einer Regel an den Leerstellen in Teilstrings. Diese
	# werden dem Konstruktor der Klasse Symb bergeben und die neuen
	# Symbole in das Array @symbs gesteckt.
	def initialize(alt)
		@symbs = Array.new
		alt.split(/\s*/).each do |symb|
			@symbs.push(Symb.new(symb))
		end
	end
	
	# Einfacher Getter fr @symbs (Array aller Symbole)
	def getSymbols
		return @symbs
	end
	
	# Gibt die Anzahl der Symbole der Alternative zurck
	def countSymbols
		return @symbs.size
	end
	
	# Operator der den Zugriff auf die einzelnen Symbole der Alternative
	# ermglicht. alternative[i] liefert das Symbol das im @symbs-Array
	# an Index i liegt.
	def [] (ind)
		return @symbs[ind]
	end
end

# Diese Klasse reprsentiert die Einzelnen Regeln. Jede Regel besteht aus 
# einem Symbol als linker Regelseite und einem Array mit Alternativen als 
# linker Seite der Regl.
class Rule
	
	# Der Konstruktor einer Regel bekommt einen String mit der Kompletten
	# Regel bergeben. Dieser wird am Trennzeichen -> in die linke und rechte
	# Seite geteilt. Die linke Seite wird dem Konstruktor von Symb bergeben
	# und das neue Symbol im @symbol-ATtribut abgelegt. Die rechte Seite wird
	# weiter zerlegt in die verschiedenen Alternativen (Nach Zeilenumbruch
	# beliebiger EInrckung und dem Zeichen |). Diese werden wiederum
	# dem Konstruktor von Alternative bergeben und die so entstehenden 
	# Alternativen in das @alts-Array gepackt.
	def initialize(rule)
		leftside, rightside = rule.split(/\s*->\s*/)
		@symbol = Symb.new(leftside)
		@alts = Array.new
		rightside.split(/\s*\n\s*\|\s*/).each do |a|
			@alts.push Alternative.new(a)
		end	
	end
	
	# Einfacher Getter fr @symbol
	def getSymbol
		return @symbol
	end
	
	# Einfacher Getter fr @alts (Array aller Alternativen)
	def getAlts
		return @alts
	end
	
	# Gibt die Anzahl der Alternativen einer Regel zurck
	def countAlts
		return @alts.size
	end
	
	# Operator der den Zugriff auf die einzelnen ALternativen der Regel
	# ermglicht. regel[i] liefert das Symbol das im @alts-Array
	# an Index i liegt. 
	def [] (ind)
		return @alts[ind]
	end
end

# Diese Klasse reprsentiert eine Komplette Grammatik. Sie hat ein ein 
# Atribut fr das Startsymbol und jeweils ein Hash-wertiges Attribut fr
# die Teminale, Non-Terminale und die Regeln. Die Schlssel fr (Non-)
# Terminale sind die Namen der Symbole und fr die Regeln das Non-Terminal
# welches auf der linken Seite der Regel steht.
class Grammar
	
	# Der Konstruktor der Grammatik bekommtden Regelteil einer SIC-Grammatik
	# als String bergeb und spaltet ihn an einer Leerzeile in die einzelnen
	# Regel, welche wiederum an den Konstruktor von Rule bergeben wird. Die
	# so entstehenden Reglen kommen in den @rules-Hash. Das Symbol der ersten
	# Regel wird das Startsymbol der Grammatik, alle anderen linksseitigen
	# Symbole die Non-Terminale. Die Terminale ergeben sich aus allen Symbolen
	# der zu allen Regeln gehrigen Alternativen die nicht vom Namen her mit
	# einem Non-Terminal bereinstimmen. 
	#
	# (Bei genauer betrachtung stellt man fest, das so immer nur ein gleichnamiges
	# Symbol in den Hash steht. Da Symbole gleichen Namens aber an fr sich gleich
	# sind strt diese Verwerfung nicht.)
	def initialize(gram)
		@rules = Hash.new
		@terminals = Hash.new
		@nonterminals = Hash.new
		gram.split(/\s*\n\s*\n\s*/).each do |rule|
			newrule = Rule.new(rule)
			if @rules.size == 0
				@startsymbol = newrule.getSymbol
			end
			@rules[newrule.getSymbol] = newrule
			@nonterminals[newrule.getSymbol.getName] = newrule.getSymbol
		end
		@rules.each_value  {|rule|
			rule.getAlts.each do |alt|
				alt.getSymbols.each do |symb|
					if @nonterminals.has_key?(symb.getName) == false
						@terminals[symb.getName] = symb
					end
				end
			end
		}
	end
	
	# Einfacher Getter fr @startsymbol
	def getStartsymbol
		return @startsymbol
	end
	
	# Einfacher Getter fr @terminals (Hash aller Terminale)
	def getTerminals
		return @terminals
	end
	
	# Einfacher Getter fr @nonterminals (Hash aller Non-Terminale)
	def getNonterminals
		return @nonterminals
	end
	
	# Einfacher Getter fr @rules (Hash aller Regeln)
	def getRules
		return @rules
	end
end
