require "Warshall.rb"
require "Grammatiken.rb"

# Author::    Jens Bachmann  (mailto:jens.bachmann@unibw.de)
# Author::		Daniel Borkowitz (mailto:daniel.borkowitz@unibw.de)
# Author::		Stefan Mayer (mailto:stefan.mayer@unibw.de)
# Author::		Roman Krhling (mailto:roman.kraehling@unibw.de)
# Copyright:: Copyright (c) 2006 The Heimschlfers, UniBw Mnchen

# Diese Klasse stellt einen Zustand eines Item-Automaten(NEA) dar, z.B. [S->X.Ya]. (S.154)
class Zustand
	
	def initialize(string)
	@item=string
	end
	
	def getZustand
	 return @item
	end
	
	def +(it)
	   Zustand.new(@item + it.getZustand)
	
	end
	
	def each &b
		@item.each &b
	end
	
	
	
	def item!(str)
		@item = str
	end
	# setzt den Punkt um ein Zeichen nach rechts weiter
	def lesen
		ind = getZustand.index('.')
		ch = @item[ind+1..ind+1]
		item= Zustand.new("")
		item+= Zustand.new(@item[0..ind-1]+ch+"."+@item[ind+2..@item.length])
	end 
 
	def <=>(anOther)
		@item<=>anOther.getZustand
	end
	
	def ==(anOther)
		return @item == anOther.getZustand if anOther!=nil
		return false
	end
	
	def sort!()
	@item.sort!
	end
	
	def to_s
		"[" + @item + "]"
	end
	
	# liefert das nchste Zeichen nach dem Punkt
	def naechstesZeichen
		ind = getZustand.index('.')
		return @item[ind+1..ind+1]
	end
	
	# prft ob das Folgezeichen des Punktes ein NonTerminal ist
	def isEpsUebergang(gram)
		return gram.getNonterminals.include?(naechstesZeichen)
	end
	
	# prft ob der Zustand der Ausgangszustand zu einem NonTerminal a ist
	def ist_start_von(a)
		ind = getZustand.index('.')		
		return ((@item[ind-1..ind-1]) == ">" and (@item[0..0] == a))
	end
	
end

# Diese Klasse stellt das Tripel einer LR-Konfiguration dar (S.155)
class Konfiguration
	def initialize(zustand, symbol, folge)
		@zustand = zustand
		@symbol = symbol
		@folge = folge
	end
	
	# getter fr die Komponenten des Tripels(0-2)
	def [](i)
	 case i
	    when 0 then @zustand
			when 1 then @symbol
			when 2 then @folge
			else nil
		end
	end
	
	def ==(anOther)
	  if anOther!=nil
			return (@zustand==anOther[0] and @symbol==anOther[1] and @folge==anOther[2]) 
		else return false
		end
	end
	
	def <=>(anOther)
		return @zustand<=>anOther[0]
	end
	
	def to_s
		return ("[" + @zustand.to_s + "," + @symbol.to_s + "," + @folge.to_s + "]" + "\n")
	end
	
	# liefert den Folgezustand eines Zustands nach "lesen" eines Symbols
	def folgeZustand(zustand,symbol)
	 return @folge if ((zustand==@zustand) and (@symbol == symbol))
   return nil
	end	
end

# Diese Klasse stellt einen Nichtdeterministischen Item-Automaten zu einer Grammatik dar.
class NEA
	
	# der NEA besteht aus einer Menge von Zustnden, einem Startzustand und einer Menge von 
	# Konfigurationen als bergangsfunktion.
	def initialize(grammatik)
			@zustaende = Menge.new([])
			@startZustand = Zustand.new("@ ->.S") 
			berechneZustaende(grammatik.addRule("@ -> S \n\n"))
			@uebergaenge = uebergangsfunktion(grammatik.addRule("@ -> S \n\n"))
 end

	
	# berechnet die Zustnde des Automaten als Menge
	def berechneZustaende(grammatik)
		grammatik.getRules.each_value do |x| 
				links = x.getSymbol
				tmpMenge = Menge.new(x.getAlts)
				tmpMenge.each do |i| 
					for b in 0..i.getSymbols.length do
						str=Zustand.new(links.to_s+" ->")
						for c in 0..i.getSymbols.length-1 do
						  
							str+=Zustand.new(".") if ((c==b))
							str+=Zustand.new(i.getSymbols[c].to_s)
						end
						str+=Zustand.new(".") if (b==i.getSymbols.length)
						@zustaende.add(str)
				end
			end
		end
	end
	
	# berechnet die bergnge des Automaten als Menge von LR-Konfigurationen
	def uebergangsfunktion(grammatik)
		erg = Menge.new([])
		@zustaende.each do |z| 
				 ch = z.naechstesZeichen
				 if (z.isEpsUebergang(grammatik)) then
				    @zustaende.each do |folg| 
						erg.add(Konfiguration.new(z,Symb.new("eps"),folg)) if  folg.ist_start_von(ch.to_s) and folg!=z
						end
					end
					erg.add(Konfiguration.new(z,Symb.new(ch),z.lesen)) if ch!=""
			end
  		return erg
	end
	
	def getUebergaenge
		@uebergaenge
	end
	
	def getZustaende
	 @zustaende
	end
	
	def getStart_Zustand
		@startZustand
	end

	# each-Iterator fr @zustaende
	def zustaende_each &b
	 @zustaende.each &b
	end
	
	# each-Iterator fr @uebergaenge
	def uebergaenge_each &b
	 @uebergaenge.each &b
	end
end



