# Author::		Sebastian Bauer
# Author::		Norman Bhm
# Author::		Alexander Jede
# Author::		Sergej Krun
# Author::		Ren Wickmann

# Grammatiken
# ===========
#
# (Vorlufige Version)
# Dieses Programm liest einen Regelteil einer SIC-Grammatik ein,
# zerlegt ihn in seine Bestandteile und stellt einige Methoden 
# zur Verfgung um mit den so entstandenen Teilen zu arbeiten.

# Diese Klasse reprsentiert die Symbole der Grammatik.
# Symbole kennen ihren Namen und knnen diesen zurckgeben.
# Und sie wissen, ob sie terminale oder nichtterminale Symbole sind.
# Diese Eigenschaft kann von aussen verndert und auch ausgegeben werden
# (Beides jeweils mit einem direktem Zugriff auf das Attribut _Symb_._terminal_ )
class Symb
    attr_accessor :terminal
	# Einfacher Konstruktor
	def initialize(name)
		@name = name
		@terminal = true
	end
	
	# Einfacher Getter fr @name-Attribut
	def getName
		return @name
	end
	##
	# Die to_s Methode gibt den Namen des Symbols als String wieder
	def to_s
	@name
	end

	def ==(anOther)
		(@name == anOther.getName) && (@terminal == anOther.terminal)
	end
end

# Diese Klasse reprsentiert die einzelne Alternative auf der
# rechten Seite einer Regel. Sie besteht aus einem Array von 
# Symbolen.
class Alternative

	# Der Konstruktor zerlegt einen bergebenen String einer rechten
	# Seite einer Regel an den Leerstellen in Teilstrings. Diese
	# werden dem Konstruktor der Klasse Symb bergeben und die neuen
	# Symbole in das Array @symbs gesteckt.
	def initialize(alt = nil)
		@symbs = Array.new
		if alt != nil
			alt.split(/\s+/).each do |symb|
				@symbs.push(Symb.new(symb))
			end		
		end
	end
	
	# Einfacher Getter fr @symbs (Array aller Symbole)
	def getSymbols
		return @symbs
	end
	
	# Gibt die Anzahl der Symbole der Alternative zurck
	def countSymbols
		return @symbs.size
	end
	
	# Operator der den Zugriff auf die einzelnen Symbole der Alternative
	# ermglicht.  alternative[i] liefert das Symbol das im @symbs-Array
	# an Index i liegt.
	def [] (ind)
		return @symbs[ind]
	end
	
	##
	# Die to_s Methode um die Alternative als String aus zugeben.
	def to_s
	res = String.new
	@symbs.each{ |symb| res = res+" "+symb.to_s }
	return res
	end
	
	##
	# Diese Methode liefert einen boolschen Wert (true oder false) zurck, ob aus dieser Alternative das Leerezeichen Epsilon erzeugt werden kann.
	# Dies ist der Fall, wenn eine Alternative nur aus Nonterminalen besteht,
	# die auch alle zu Epsilon abgeleitet werden knnen.
	# 
	def isEpsilon
		@symbs.empty?
	end
	
end

# Diese Klasse reprsentiert die einzelnen Regeln. Jede Regel besteht aus 
# einem Symbol als linker Regelseite und einem Array mit Alternativen als 
# rechter Seite der Regel.
class Rule
	
	# Der Konstruktor einer Regel bekommt einen String mit der kompletten
	# Regel bergeben. Dieser wird am Trennzeichen -> in die linke und rechte
	# Seite geteilt. Die linke Seite wird dem Konstruktor von Symb bergeben
	# und das neue Symbol im @symbol-Attribut abgelegt. Die rechte Seite wird
	# weiter zerlegt in die verschiedenen Alternativen (Nach Zeilenumbruch
	# beliebiger Einrckung und dem Zeichen |). Diese werden wiederum
	# dem Konstruktor von Alternative bergeben und die so entstehenden 
	# Alternativen in das @alts-Array gepackt.
	def initialize(rule)
		leftside, rightside = rule.split(/\s*->\s*/)
		@symbol = Symb.new(leftside)
		@symbol.terminal = false
		@alts = Array.new
		if rightside == nil
			@alts.push Alternative.new()
		else	
			rightside.split(/\n\s*\|\s*/).each do |a|
				@alts.push Alternative.new(a)
			end
		end
	end
	
	# Einfacher Getter fr @symbol
	def getSymbol
		return @symbol
	end
	
	# Einfacher Getter fr @alts (Array aller Alternativen)
	def getAlts
		return @alts
	end
	
	# Gibt die Anzahl der Alternativen einer Regel zurck
	def countAlts
		return @alts.size
	end
	
	# Operator der den Zugriff auf die einzelnen Alternativen der Regel
	# ermglicht. regel[i] liefert das Symbol das im @alts-Array
	# an Index i liegt. 
	def [] (ind)
		return @alts[ind]
	end
	
	##
	# Die to_s Methode, die eine Regel in einem String ausgibt.
	# Auf der linken Seite des Pfeiles ( "->" ) steht das Non-Terminal.
	# Rechts davon stehen dann die Alternativen, die jeweils durch einen "|" getrennt sind.
	# Jede einzelne Alternative steht in einer neuen Zeile.
	# Die einzelnen Alternativen, Trennzeichen und das linke Symbol 
	# sind durch Leerzeichen getrennt.
	def to_s
	res = @symbol.to_s+" ->"
	@alts.each_index{|i|
	  if i != @alts.length-1
	   res = res + " " + @alts[i].to_s + "\n   |"
	  else
	   res = res + " " + @alts[i].to_s
	  end
	}
	# Die Liste aller Alternativen in eine Variable speichern und das erste Element entfernen,
	# da es schon im Result steht
	res
	end
	
	##
	# Diese Methode liefert einen boolschen Wert (true oder false) zurck, ob aus dieser Regel das Leerezeichen
	# Epsilon erzeugt werden kann. Dazu muss Epsilon entweder direkt ableitbar sein, oder eine Alternative muss 
	# sich in beliebig vielen Schritten zu Epsilon ableiten lassen.
	def isEpsilon
		if @alts.empty?
			return true
		else
			@alts.each{|alt|
				if alt.isEpsilon
					return true
				end
			}
		end
		return false
	end
	
		def Rule.from_Yacc(input)
		i = 1
		output = ""
		puts input + "\n"
		zeilen = input.split("\n")
		name = zeilen[0].split[1]
		# Erste Zeile
		output += name + " -> " + name + i.to_s + "\n"
		# Mittlere Zeilen
		zeilen.delete_at(0)
		zeilen.each do |zeile|
			optoken = zeile.split[0]
			operator = zeile.split[1]
			if optoken == "left"
				output += name + i.to_s + " -> " + name + i.to_s + " " + operator + " " + name + (i+1).to_s + "\n"
				output += " " * (name + i.to_s).length + " | " + name + (i+1).to_s + "\n"
				i += 1
			elsif optoken == "right"
				output += name + i.to_s + " -> " + name + (i+1).to_s + " " + operator + " " + name + i.to_s + "\n"
				output += " " * (name + i.to_s).length + " | " + name + (i+1).to_s + "\n"
				i += 1
			elsif optoken == "none"
				output += name + i.to_s + " -> " + name + (i+1).to_s + " " + operator + " " + name + (i+1).to_s + "\n"
				output += " " * (name + i.to_s).length + " | " + name + (i+1).to_s + "\n"
				i += 1
			elsif optoken == "pre"
				output += name + i.to_s + " -> " + operator + " " + name + i.to_s + " " + "\n"
				output += " " * (name + i.to_s).length + " | " + name + (i+1).to_s + "\n"
				i += 1
			elsif optoken == "post"
				output += name + i.to_s + " -> " + name + i.to_s + " " + operator + " " + name + (i+1).to_s + "\n"
				output += " " * (name + i.to_s).length + " | " + name + (i+1).to_s + "\n"
				i += 1
			end
		end
		# Letzte Zeile
		output += name + i.to_s + " -> " + "(" + name + 1.to_s + ")" +"\n"
		atome = zeilen[zeilen.size-1].split(/,\s|\s/)
		atome.delete_at(0)
		atome.each do |atom|
			output += " " * (name + i.to_s).length + " | " + atom + "\n"
		end
		puts output
	end

end

# Diese Klasse reprsentiert eine Komplette Grammatik. Sie hat ein ein 
# Atribut fr das Startsymbol und jeweils ein Hash-wertiges Attribut fr
# die Teminale, Non-Terminale und die Regeln. Die Schlssel fr (Non-)
# Terminale sind die Namen der Symbole und fr die Regeln das Non-Terminal
# welches auf der linken Seite der Regel steht.
class Grammar
	
	# Der Konstruktor der Grammatik bekommt den Regelteil einer SIC-Grammatik
	# als String bergeb und spaltet ihn an einer Leerzeile in die einzelne
	# Regel, welche wiederum an den Konstruktor von Rule bergeben wird. Die
	# so entstehenden Regeln kommen in den @rules-Hash. Das Symbol der ersten
	# Regel wird das Startsymbol der Grammatik, alle anderen linksseitigen
	# Symbole sind die Non-Terminale. Die Terminale ergeben sich aus allen Symbolen
	# der zu allen Regeln gehrigen Alternativen die nicht vom Namen her mit
	# einem Non-Terminal bereinstimmen. 
	#
	# (Bei genauer Betrachtung stellt man fest, dass so immer nur ein gleichnamiges
	# Symbol in dem Hash steht. Da Symbole gleichen Namens aber an und fr sich gleich
	# sind strt diese Verwerfung nicht.)
	def initialize(gram)
		@rules = Hash.new
		@terminals = Hash.new
		@nonterminals = Hash.new
		gram.split(/\s*\n\s*\n\s*/).each do |rule|
			newrule = Rule.new(rule)
			if @rules.size == 0
				@startsymbol = newrule.getSymbol
			end
			@rules[newrule.getSymbol] = newrule
			@nonterminals[newrule.getSymbol.getName] = newrule.getSymbol
		end
		@rules.each_value  {|rule|
			rule.getAlts.each do |alt|
				alt.getSymbols.each do |symb|
					if @nonterminals.has_key?(symb.getName) == false
						@terminals[ symb.getName ]= symb
					end
				end
			end
		}
	end
	
	# Einfacher Getter fr @startsymbol
	def getStartsymbol
		return @startsymbol
	end
	
	# Einfacher Getter fr @terminals (Hash aller Terminale)
	def getTerminals
		return @terminals
	end
	
	# Einfacher Getter fr @nonterminals (Hash aller Non-Terminale)
	def getNonterminals
		return @nonterminals
	end
	
	# Einfacher Getter fr @rules (Hash aller Regeln)
	def getRules
		return @rules
	end
	
	##
	# Die to_s Methode liefert die Grammatik in einem String zurck.
	# Dabei steht jede Regel in einer Zeile und zwischen den Regeln ist eine Leerzeile.
	# Fr das Format der Regeln, siehe Rule#to_s
	def to_s
	res = @rules[@startsymbol].to_s
	@rules.each_pair{|key,r|
	 if key != @startsymbol
	   res = res + "\n\n" +r.to_s
	 end
	}
	res
	end
	
	##
	#
	#
	
	def epsilon_Symbole
		res = Array.new
		#Es mssen alle Regeln durchsucht werden, ob sie zu epsilon werden
		# Solange Epsilon Produktionen vorhanden sind, eine Schleife durchlaufen,
		# die nach epsilon-Regeln sucht
		keine_epsilon = false
		while !keine_epsilon
			keine_epsilon = true
			# Suchen, ob noch eine Epsilon-Ableitung vorhanden is
			@rules.each_pair {|name,rule|
				# Nachschauen ob jede Regel zu Epsilon wird
				if rule.isEpsilon
					res.include?(name) ? () : (res.push(name);kein_epsilon = false)
				end
				#else
					# Nach schauen ob eine Alternative sich NUR aus den Symbolen zusmmenstellen
					# lsst, die sich zu Epsilon ableiten lassen
					rule.getAlts.each {|alt|
						abbruch =false
						# Eine Alternative dann Epsilon, wenn sie nur aus NonTerminalen besteht,
						# die zu epsilon werden
						alt.getSymbols.each{ |symb|
							res.include?(symb) ? () : (abbruch = true;) 
						}
						if !abbruch
							res.include?(name) ? () : (res.push(name); keine_epsilon = false)
						end
					}
					
				#end
			}
		end
		return res
	end

	
	##
	# Diese Methode ermglicht das Speichern der Grammatik in eine Datei.
	# Als Format wird YAML ( http://www.yaml.org/ ) benutzt, somit ist die Datei im Plaintext und kann auch mit einem
	# einfachem Editor bearbeitet werden. *Vorsicht*: Dabei kann man die Formatierung beschdigen, so dass
	# die gespeicherte Grammatik nicht mehr geladen werden kann.
	# file - Der Pfad zu der Datei in Form eines Strings, in der die Grammatik gespeichert werden soll.
	# Sollte die Datei nicht vorhanden sein, wird sie erstellt. <b> Ansonsten wird die Datei berschrieben! </b>
	def save(file)
	 require 'yaml'
	 gramma = Grammar.new(to_s)
	 File.open(file,"w"){|data|
	   YAML.dump(gramma,data)
	 }
	end
	
	##
	# Diese Methode ermglicht es eine in einer Datei abgespeicherte Grammatik wieder zu laden.
	# Ein mgliches vorgehen wre, in dem man eine leere Grammar erstellt und davon die Methode Grammar#load_grammar
	# aufruft. Zum Beispiel::
	#  g = Grammar.new("")
	#  g.load("/home/alex/SavedGrammar.yaml")
	# Grammar#load(file) - Der Pfad zu der Datei in Form eines String, in der die Grammatik gespeichert ist.
	#
	def load(file)
	 require 'yaml'
	 gramma = YAML::load(File.open(file))
	 if gramma.class == Grammar
	   @rules = gramma.getRules
	   @terminals = gramma.getTerminals
	   @nonterminals = gramma.getNonterminals
	   @startsymbol = gramma.getStartsymbol
	 else
	   #raise "Wrong File.\n It was not a Grammar"
	 end
	end
	
	##
	# Diese Methode stellt eine SIC-compatible Grammatik in Form eines Strings her.
	# Zu Beginn werden die terminalen Zeichen aufgelistet, gefolgt von den nichtterminalen Zeichen,
	# dem Startsymbol und zum Schluss den Ableitregeln.
	# Vor jedem Abschnitt leitet eine berschrift die jeweiligen Zeichen ein, der noch eine zustzliche 
	# Leerzeile folgt. 
	def to_SIC
	 res = "Terminals:\n\n"
	 getTerminals.each_value{ |termsym|
	   res += termsym.to_s + " "
	 }
	 res += "\n\nNonterminals:\n\n"
	 getNonterminals.each_value{ |nonterm|
	   res += nonterm.to_s + " " 
	 }
	 res += "\n\nStartingsymbol:\n\n" + getStartsymbol.to_s
	 res += "\n\nRules:\n\n" + to_s
	 return res	 
	end
end

puts "P -> G \n   |H \n\n D -> P P"
g = Grammar.new("P -> G \n   |H \n\nD -> P P")
puts g.epsilon_Symbole
g2 = Grammar.new("S -> A P\n  | a P\n\nA -> P\n  | b A\n\nP -> ")
puts g2.to_s
puts g2.epsilon_Symbole
puts "--------------------"
#puts "Die Grammatik"
#puts g.to_s
#g.save("Test.yaml")
#t = Grammar.new("")
#t.load("Test.yaml")
#puts "----- t: ------"
#puts t.to_s
#puts "------ to_SIC---------"
#puts g.to_SIC

#Testflle der Klasse Symb
require "test/unit"
class SymbToTest < Test::Unit::TestCase

	def setup
	   @symbol1 = Symb.new("s")
	   @symbol2 = Symb.new("T")
	   @symbol2.terminal=false
	end
	
	def test_getName
	   assert("s" == @symbol1.getName)
	   assert("T" == @symbol2.getName)
	end
	
	def test_to_s
	   assert("T" == @symbol2.to_s)
	end
	
	def test_attr_accessor
	   assert(@symbol1.terminal)
	   assert(!(@symbol2.terminal))
	end
end

#Testflle der Klasse Alternative
class AlternativeToTest < Test::Unit::TestCase

	def setup
	   @alternative1 = Alternative.new("H")
	   @alternative2 = Alternative.new("A B c")
	   @alternative3 = Alternative.new(" ")
	   @symbol3 = Symb.new("B")
	end
	
	def test_getSymbols
	   # to_s ist ntig, da die Adressen der beiden nicht bereinstimmen
	   assert("H" == @alternative1.getSymbols.to_s)
	   assert_equal("ABc", @alternative2.getSymbols.to_s)
	   assert_equal(@symbol3.to_s, @alternative2.getSymbols[1].to_s)
	end
	
	def test_countSymbols
	   assert(1 == @alternative1.countSymbols)
	   assert(3 == @alternative2.countSymbols)
	end
	
	def test_ind
	   assert_equal("H", @alternative1.[](0).to_s)
	   assert_equal("c", @alternative2.[](2).to_s)
	   assert_equal("", @alternative1.[](2).to_s)
	end
	
	def test_to_s
	   assert(" H" == @alternative1.to_s)
	   assert(" A B c" == @alternative2.to_s)
	end
	
	def test_isEpsilon
	   assert(!(@alternative1.isEpsilon))
	   assert(!(@alternative2.isEpsilon))
	   #noch ein Test wo ein Epsilon dabei ist
	   assert(@alternative3.isEpsilon)
	end
end

#Testflle der Klasse Rule
class RuleToTest < Test::Unit::TestCase

	def setup
	   @regel1 = Rule.new("P -> g \n   |h") 
	   @regel2 = Rule.new("D -> p p")
	end
	
	def test_getSymbol
	   assert_equal("P", @regel1.getSymbol.to_s)
	   assert(!(@regel1.getSymbol.terminal))  # mit "!", da die linke Regelseite ein Non-Terminal ist
	end
	
	def test_getAlts
	   assert_equal(" g h", @regel1.getAlts.to_s)
	   assert_equal(" p p", @regel2.getAlts.to_s)
	   assert_equal(" p p", @regel2.getAlts[0].to_s)
	   assert_equal(nil, @regel2.getAlts[1])
	end
	
	def test_countAlts
	   assert(2 == @regel1.countAlts)
	   assert(1 == @regel2.countAlts)
	end
	
	def test_ind
	   assert_equal(" h", @regel1.[](1).to_s)
	end
	
	def test_to_s
	   assert_equal("P ->  g\n   |  h", @regel1.to_s)
	   assert_equal("D ->  p p", @regel2.to_s)
	end
	
	def test_isEpsilon
	   assert(!(@regel1.isEpsilon))
	   assert(!(@regel2.isEpsilon))
	   #noch ein Test fr eine Epsilonproduktion
	end
end

#Testflle der Klasse Grammar
class GrammarToTest < Test::Unit::TestCase

	def setup
	   @grammatik1 = Grammar.new("P -> g \n   |h \n\nD -> p p")
	   @grammatik2 = Grammar.new("S -> H \n   |h \n\nH -> A \n   |B \n   |c \n\nA -> a \n   |b \n\nB -> ")
	   @regel3 = Rule.new("P -> g \n   |h") 
	   @regel4 = Rule.new("D -> p p")
	end
	
	def test_getStartsymbol
	   assert_equal("P",@grammatik1.getStartsymbol.to_s)
	   assert(!(@grammatik1.getStartsymbol.terminal))
	   assert_equal("S",@grammatik2.getStartsymbol.to_s)
	   assert(!(@grammatik2.getStartsymbol.terminal))
	end
	
	def test_getTerminals
	   #puts @grammatik1.getTerminals
	   #assert_equal("ghp", @grammatik1.getTerminals.to_s)
	   #assert_equal("hcab", @grammatik2.getTerminals.to_s)
	end
	
	def test_getNonterminals
	   #puts @grammatik1.getNonterminals
	   #assert_equal("PD", @grammatik1.getNonterminals.to_s)
	   #assert_equal("SHAB", @grammatik1.getNonterminals.to_s)
	end
	
	def test_getRules
	   #puts @grammatik1.getRules.to_s
	   #assert_equal(@regel3, @grammatik1.getRules.[](0) )
	end
	
	def test_to_s
	   assert_equal("P ->  g\n   |  h\n\nD ->  p p", @grammatik1.to_s)
	end
	
	def test_to_SIC
	  assert_equal("Terminals:\n\np g h \n\nNonterminals:\n\nD P \n\nStartingsymbol:\n\nP\n\nRules:\n\nP ->  g\n   |  h\n\nD ->  p p", @grammatik1.to_SIC)
	end
end

#Rule.from_Yacc("name Arith\nleft addOp\nleft multOp\nright expOp\npre addOp\natom zahl, name\n")
#Rule.from_Yacc("name L\nnone logBinOp\npre negOp\natom true, false, name\n")
