require "B_syntaxbume.rb"

class Rule

	def initialize(alt)
		@alt = Array.new
	end
	
# Die Klassenmethode Rule.from_Yacc erzeugt aus den bergebenen Kurzangaben
# eine Ausdrucksgrammatik
# Genaue Beschreibung siehe "Grammatiken".
#
# Beispiel:
# Kurzangaben zu den arithmetischen Ausdruecken:
#
# name Arith
# left addOp
# left multOp
# right expOp
# pre addOp
# atom zahl, name
#
# Grammatik zu der konkreten Syntax:
#
# Arith -> Arith1
# Arith1 -> Arith1 addOp Arith2
#        | Arith2
# Arith2 -> Arith2 multOp Arith3
#        | Arith3
# Arith3 -> Arith4 expOp Arith3
#        | Arith4
# Arith4 -> addOp Arith4
#        | Arith5
# Arith5 -> (Arith1)
#        | zahl
#        | name

	def Rule.from_Yacc(input)
		i = 1
		output = ""
		zeilen = input.split("\n")
		name = zeilen[0].split[1]
		# Erste Zeile
		output += name + " -> " + name + i.to_s + "\n"
		# Mittlere Zeilen
		zeilen.delete_at(0)
		zeilen.each do |zeile|
			optoken = zeile.split[0]
			operator = zeile.split[1]
			if optoken == "left"
				output += name + i.to_s + " -> " + name + i.to_s + " " + operator + " " + name + (i+1).to_s + "\n"
				output += " " * (name + i.to_s).length + " | " + name + (i+1).to_s + "\n"
				i += 1
			elsif optoken == "right"
				output += name + i.to_s + " -> " + name + (i+1).to_s + " " + operator + " " + name + i.to_s + "\n"
				output += " " * (name + i.to_s).length + " | " + name + (i+1).to_s + "\n"
				i += 1
			elsif optoken == "none"
				output += name + i.to_s + " -> " + name + (i+1).to_s + " " + operator + " " + name + (i+1).to_s + "\n"
				output += " " * (name + i.to_s).length + " | " + name + (i+1).to_s + "\n"
				i += 1
			elsif optoken == "pre"
				output += name + i.to_s + " -> " + operator + " " + name + i.to_s + " " + "\n"
				output += " " * (name + i.to_s).length + " | " + name + (i+1).to_s + "\n"
				i += 1
			elsif optoken == "post"
				output += name + i.to_s + " -> " + name + i.to_s + " " + operator + " " + name + (i+1).to_s + "\n"
				output += " " * (name + i.to_s).length + " | " + name + (i+1).to_s + "\n"
				i += 1
			end
		end
		# Letzte Zeile
		output += name + i.to_s + " -> " + "(" + name + 1.to_s + ")" +"\n"
		atome = zeilen[zeilen.size-1].split(/,\s|\s/)
		atome.delete_at(0)
		atome.each do |atom|
			output += " " * (name + i.to_s).length + " | " + atom + "\n"
		end
		output
	end
	
# Die Klassenmethode Rule.from_Yacc2 erzeugt aus den bergebenen Kurzangaben
# eine Ausdrucksgrammatik OHNE Przedenzen und Klammern
#
# Beispiel:
# Kurzangaben zu den arithmetischen Ausdruecken:
#
# name Arith
# left addOp
# left multOp
# right expOp
# pre addOp
# atom zahl, name
#
# Grammatik zu der abstrakten Syntax:
#
# Arith -> Arith addOp Arith
#       | Arith multOp Arith
#       | Arith expOp Arith
#       | addOp Arith
#       | zahl
#       | name
#

	def Rule.from_Yacc2(input)
		output = ""
		zeilen = input.split("\n")
		name = zeilen[0].split[1]
		zeilen.delete_at(0)
		output += name + " -> "
		# Erzeugung der Alternativen
		zeilen.each do |zeile|
			optoken = zeile.split[0]
			operator = zeile.split[1]
			if (optoken == "left" || optoken == "right" || optoken == "none")
				output += name + " " + operator + " " + name + "\n" + " " * name.length + " | "
			elsif optoken == "pre"
				output += operator + " " + name + " " + "\n" + " " * name.length + " | "
			elsif optoken == "post"
				output += name + " " + operator + " " + "\n" + " " * name.length + " | "
			end
		end
		# Letzte Zeile
		atome = zeilen[zeilen.size-1].split(/,\s|\s/)
		atome.delete_at(0)
		atome.each do |atom|
			output += atom + "\n" + " " * name.length + " | "
		end
		# Korrektur der letzten Zeile
		output.strip.chop.strip
	end

end

###################################################################################################
puts "Kurzangaben zu den arithmetischen Ausdruecken:\n"
puts "\n"
puts kurzangaben1 = "name Arith\nleft addOp\nleft multOp\nright expOp\npre addOp\natom zahl, name\n"
puts "\n"
puts "Grammatik zu der konkreten Syntax --> Rule.from_Yacc(kurzangaben):\n"
puts "\n"
puts Rule.from_Yacc(kurzangaben1)
puts "\n"
puts "Grammatik zu der abstrakten Syntax --> Rule.from_Yacc2(kurzangaben):\n"
puts "\n"
puts Rule.from_Yacc2(kurzangaben1)
puts "\n"

=begin
puts kurzangaben2 = "name L\nnone logBinOp\npre negOp\natom true, false, name\n"
puts "\n"
puts Rule.from_Yacc(kurzangaben2)
puts "\n"
puts Rule.from_Yacc2(kurzangaben2)
puts "\n"
=end
###################################################################################################

# Die Methode reduce(baum) transformiert den bergebenen Baum in einen Baum
# ohne Kettenproduktionen. Dieser Baum entspricht der abstrakten Grammatik.
# Zustzlich werden alle Nonterminale umbenannt und bekommen die Bezeichnung
# der Wurzel.
#
# Beispiel:
# Syntaxbaum zu 1 + 2 bzgl. der konkreten Syntax:
#
# Arith
#  Arith1
#   Arith1
#    Arith2
#     Arith3
#      Arith4
#       Arith5
#        1
#   +
#   Arith2
#    Arith3
#     Arith4
#      Arith5
#       2
#
# Syntaxbaum zu 1 + 2 bzgl. der abstrakten Syntax:
#
# Arith
#  Arith
#   1
#  +
#  Arith
#   2

def reduce(baum)
	if baum != nil && baum.is_a?(Nonterminal)
		# Fall 1: Kettenproduktion und ...
		if baum.getSoehne().size == 1
			# Fall 1a: ... es liegt ein Nonterminal vor -> kann "reduziert" werden
			if baum.getSoehne()[0].is_a?(Nonterminal)
				name = baum.getName()
				baum.setSoehne(baum.getSoehne()[0].getSoehne())
				baum.getSoehne().each do |sohn|
					sohn.setVater(baum)
					sohn.setName(name)
				end
				reduce(baum)
			# Fall 1b: ... es liegt ein Terminal vor -> nichts zu tun
			end
		# Fall 2: keine Kettenproduktion
		else
			baum.getSoehne().each do |sohn|
				reduce(sohn)
			end
		end
	end
end

testbaum = Syntaxbaum.new("Arith",0)
testbaum.addNonterminal("Arith1",1)
testbaum.getSoehne()[0].addNonterminal("Arith1",2)
testbaum.getSoehne()[0].addTerminal("+","+")
testbaum.getSoehne()[0].addNonterminal("Arith2",3)
testbaum.getSoehne()[0].getSoehne()[0].addNonterminal("Arith2",4)
testbaum.getSoehne()[0].getSoehne()[0].getSoehne()[0].addNonterminal("Arith3",5)
testbaum.getSoehne()[0].getSoehne()[0].getSoehne()[0].getSoehne()[0].addNonterminal("Arith4",6)
testbaum.getSoehne()[0].getSoehne()[0].getSoehne()[0].getSoehne()[0].getSoehne()[0].addNonterminal("Arith5",7)
testbaum.getSoehne()[0].getSoehne()[0].getSoehne()[0].getSoehne()[0].getSoehne()[0].getSoehne()[0].addTerminal("1","1")
testbaum.getSoehne()[0].getSoehne()[2].addNonterminal("Arith3",8)
testbaum.getSoehne()[0].getSoehne()[2].getSoehne()[0].addNonterminal("Arith4",9)
testbaum.getSoehne()[0].getSoehne()[2].getSoehne()[0].getSoehne()[0].addNonterminal("Arith5",10)
testbaum.getSoehne()[0].getSoehne()[2].getSoehne()[0].getSoehne()[0].getSoehne()[0].addTerminal("2","2")
puts "Syntaxbaum zu 1 + 2 bzgl. der konkreten Syntax:\n\n"
testbaum.drucken

puts "\n\n"
puts "Syntaxbaum zu 1 + 2 bzgl. der abstrakten Syntax:\n\n"

tbaum = reduce(testbaum)
testbaum.drucken

puts "\n\n"

testbaum2 = Syntaxbaum.new("Arith",0)
testbaum2.addNonterminal("Arith1",1)
testbaum2.getSoehne()[0].addNonterminal("Arith2",2)
testbaum2.getSoehne()[0].getSoehne()[0].addNonterminal("Arith3",3)
testbaum2.getSoehne()[0].getSoehne()[0].getSoehne()[0].addNonterminal("Arith4",4)
testbaum2.getSoehne()[0].getSoehne()[0].getSoehne()[0].getSoehne()[0].addNonterminal("Arith5",5)
testbaum2.getSoehne()[0].getSoehne()[0].getSoehne()[0].getSoehne()[0].getSoehne()[0].addTerminal("a","a")
puts "Syntaxbaum zu a bzgl. der konkreten Syntax:\n\n"
testbaum2.drucken

puts "\n\n"
puts "Syntaxbaum zu a bzgl. der abstrakten Syntax:\n\n"

tbaum2 = reduce(testbaum2)
testbaum2.drucken