# Die folgenden drei Klassen stellen die Testklassen fr die Datei actiontables.rb und Parser.rb dar.
# Hierbei handelt es sich um Tests fr die drei Klassen UTable, ATable und Parse sowie deren Methoden.
#  Fr die Testklassen werden zwei Grammatiken angelegt:
#   grammar1: S -> b S d | b S c S d | a
#   und
#   grammar2: E -> E + T  | T 
#             T -> T * id | id 
#  Von beiden wird ein LR(0)-Automat erstellt.
#  Von grammar2 werden noch der LR(1)- und der SLR(1)-Automat erstellt.
#
# Author:: Wickmann, Ren

require "test/unit"
require "actiontables"
require "Automaten"
require "Grammatiken"
require "Parser"
require "Warshall"


@@grammar1 = Grammar.new("S -> b S d\n | b S c S d \n | a")
@@grammar2 = Grammar.new("E -> E + T \n |T \n\n T -> T * id \n |id \n\n")
@@lr01 =  LR_0.new(@@grammar1)
@@lr02 =  LR_0.new(@@grammar2) 
@@lr1  =  LR_1.new(@@grammar2)
@@slr1 = SLR_1.new(@@grammar2)


# Diese Klasse testet die Klasse ATable.
class ATableToTest < Test::Unit::TestCase
  
  # Es werden fr alle 4 Automaten die Aktionstafeln erstellt.
  def setup 
    @atable1 = ATable.new @@lr01
    @atable2 = ATable.new @@lr02
    @atable3 = ATable.new @@lr1
    @atable4 = ATable.new @@slr1
  end
  
  # Die Methode testet, ob Konflikte in den Automaten auftauchen.
  # 1. Es wird geprft, ob beim LR(0)-Automaten von grammar1 kein Konflikt auftritt.
  # 2. Es wird geprft, ob beim LR(0)-Automaten von grammar2 3 Konflikte auftreten.
  def test_getConflicts
    assert(@atable1.getConflicts.empty? == true)
    assert_equal(@atable2.getConflicts.size, 3)
  end
  
  # Die Methode testet die Methode isConflictfree.
  # 1. Es wird geprft, ob der LR(0)-Automat von grammar1 konfliktfrei ist.
  # 2. Es wird geprft, ob der LR(0)-Automat von grammar2 nicht konfliktfrei ist.
  # 3. Es wird geprft, ob der LR(1)-Automat von grammar2 konfliktfrei ist.
  # 4. Es wird geprft, ob der SLR(1)-Automat von grammar2 konfliktfrei ist.
  def test_isConflictfree
    assert(@atable1.isConflictfree)
    assert(!(@atable2.isConflictfree))
    assert(@atable3.isConflictfree)
    assert(@atable4.isConflictfree)
  end
  
  # Die Methode testet die Mathode printConflicts.
  # 1. Es wird geprft, ob der Satz "Der Automat ist konfliktfrei." beim LR(0)-Automat von grammar1 ausgegeben wird.
  # 2. Es werden die Aktionstafeln und die Konfliktmeldungen der einzelnen Automaten ausgegeben.
  def test_printConflicts
    assert_equal("Der Automat ist konfliktfrei.", @atable1.printConflicts)
    puts
    puts @atable1
    puts
    puts @atable1.printConflicts
    puts
    puts @atable2
    puts
    puts @atable2.printConflicts
    puts
    puts @atable3
    puts
    puts @atable3.printConflicts
    puts
    puts @atable4
    puts
    puts @atable4.printConflicts
    puts
  end
end

# Diese Klasse testet die Klasse UTable
class UTableToTest < Test::Unit::TestCase

  # Es werden fr alle 4 Automaten die bergangstafeln erstellt.
  def setup
    @utable1 = UTable.new @@lr01
    @utable2 = UTable.new @@lr02
    @utable3 = UTable.new @@lr1
    @utable4 = UTable.new @@slr1
  end
  
  # Die Methode testet die Methode getnamehs.
  # 1. Es wird geprft, ob fr den LR(0)-Automaten von grammar1 9 Zustandsnamen vergeben werden.
  # 2. Es wird geprft, ob fr den LR(0)-Automaten von grammar2 8 Zustandsnamen vergeben werden.
  # 3. Es wird geprft, ob bei den sortieren Schlssesln an 4. Stelle "K5" steht.
  # 4. Es wird geprft, ob fr die Schlssel des LR(0)-Automaten von grammar2 die Namen "K1" - "K8" vergeben werden.
  # 5. Es wird geprft, ob bei sortierten bergngen, der letzte des LR(0)-Automaten von grammar2 die Form "{ [T -> id .] }" hat.
  def test_getnamehs
    assert(@utable1.getnamehs.keys.size == 9)
    assert(@utable2.getnamehs.keys.size == 8)
    assert_equal(@utable1.getnamehs.keys.sort[4], "K5")
    8.times do |i|
      assert_equal(@utable2.getnamehs.keys.sort[i], "K" + (i+1).to_s)
    end
    assert_equal(@utable2.getnamehs.values.sort.last, "{ [T -> id .] }")
  end
  
  # Die Methode testet die Methode getsymbole.
  # 1. Es werden zwei Testarrays mit den Symbolen der Grammatiken und dem #-Symbol angelegt.
  # 2. Es wird geprft, ob die Methode dieselben Werte liefert, wie die in den Testarrays erwarteten.
  # 3. Es wird das Symbol "S" und das Symbol "id" neu angelegt.
  # 4. Es wird geprft, ob das Symbol "S" dem 2. Schlssel des sortierten Symbolhashes des LR(0)-Automaten von grammar1 entspricht.
  # 5. Es wird geprft, ob das Symbol "id" dem letzten Schlssel des sortieren Symbolhashes des LR(0)-Automaten von grammar2 entspricht.
  def test_getsymbole
    testarray1 = ["#" ,"S", "a", "b", "c", "d"]
    testarray2 = ["#", "*", "+", "E", "T", "id"]
    assert_equal(@utable1.getsymbole.values.sort, testarray1.sort)
    assert_equal(@utable2.getsymbole.values.sort, testarray2.sort)
    symbol1 = Symb.new("S")
    symbol2 = Symb.new("id")
    assert_equal(@utable1.getsymbole.keys.sort[1], symbol1)
    assert_equal(@utable2.getsymbole.keys.sort.last, symbol2)
  end
  
  # Die Methode testet die Methode gettabelle.
  # 1. Es wird geprft, ob der LR(0)-Automat von grammar1 9 Zustnde hat.
  # 2. Es wird geprft, ob der LR(0)-Automat von grammar2 8 Zustnde hat.
  # 3. Es wird geprft, ob fr den Zustand K8 des LR(0)-Automaten von grammar1 keine bergnge definiert sind.
  def test_gettabelle
    9.times do |i|
      assert_equal(@utable1.getnamehs.keys.sort[i], "K" + (i+1).to_s)
    end
    assert_equal(@utable2.gettabelle.keys.size, 8)
    assert_equal(@utable1.gettabelle["K8"], {})
  end
  
  # Diese Testmethode gibt die 4 erstellten bergangstafeln aus.
  def test_to_s
    puts @utable1.to_s
    puts @utable2.to_s
    puts @utable3.to_s
    puts @utable4.to_s
  end
end

# Diese Klasse testet die Klasse Parser der Datei Parser.rb.
class ParserToTest < Test::Unit::TestCase

  # Es wird ein Parser fr den LR(0)-Automat von grammar1 und grammer2 erzeugt.
  def setup
    @parser1 = Parser.new @@lr01
    @parser2 = Parser.new @@lr1
  end
  
  # Die Methode testet die Methode parse.
  # 1. Es wird geprft, ob die Zeichenkette "bad" vom Parser akzeptiert wird.
  # 2. Es wird geprft, ob die Zeichenkette "abd" nicht vom Parser akzeptiert wird.
  # 3. Es wird geprft, ob die Zeichenkette "bbadcad" vom Parser akzeptiert wird.
  # 4. Es wird geprft, ob die Zeichenkette "id*id+id" vom Parser akzeptiert wird.
  # 5. Es wird geprft, ob die Zeichenkette "idid" nicht vom Parser akzeptiert wird.
  def test_parse
    assert(@parser1.parse("bad") != nil)
    assert_equal(@parser1.parse("abd"), nil)
    assert(@parser1.parse("bbadcad") != nil)
    assert(@parser2.parse("id*id+id") != nil)
    assert_equal(@parser2.parse("idid"), nil)
  end
  
  # Die Testmethode gibt die Syntaxbume fr die Zeichenfolgen "bad" und "bbadcad" der grammar1 sowie
  # den Syntaxbaum fr die Zeichenfolge "id*id+id" aus.
  def test_to_SB
    @parser1.to_SB("bad").drucken
    @parser1.to_SB("bbadcad").drucken
    @parser2.to_SB("id*id+id").drucken
  end
end